@@ -51,18 +51,7 @@
 
 #include <osmocom/netif/sctp.h>
 
-
-/*! \addtogroup stream Osmocom Stream Socket (client side)
- *  @{
- *
- *  This code is intended to abstract any use of stream-type sockets,
- *  such as TCP and SCTP.  It offers client side implementation,
- * fully integrated with the libosmocore select loop abstraction.
- */
-
-/*! \file stream_cli.c
- *  \brief Osmocom stream socket helpers (client side)
- */
+/*! \file stream_cli.c */
 
 #define LOGSCLI(cli, level, fmt, args...) \
 	LOGP(DLINP, level, "CLICONN(%s,%s){%s} " fmt, \
@@ -113,58 +102,34 @@ struct osmo_stream_cli {
 	uint16_t			local_port;
 	int				sk_domain;
 	int				sk_type;
+	int				sk_prio; /* socket priority, SO_PRIORITY, default=0=unset */
 	uint16_t			proto;
-	int (*connect_cb)(struct osmo_stream_cli *cli);
-	int (*disconnect_cb)(struct osmo_stream_cli *cli);
-	int (*read_cb)(struct osmo_stream_cli *cli);
-	int (*iofd_read_cb)(struct osmo_stream_cli *cli, struct msgb *msg);
-	int (*write_cb)(struct osmo_stream_cli *cli);
-	int (*segmentation_cb)(struct msgb *msg);
+	uint8_t				ip_dscp; /* IP Differentiated services, 0..63, default=0=unset */
+	osmo_stream_cli_connect_cb_t	connect_cb;
+	osmo_stream_cli_disconnect_cb_t	disconnect_cb;
+	osmo_stream_cli_read_cb_t	read_cb;
+	osmo_stream_cli_read_cb2_t	iofd_read_cb;
+	osmo_stream_cli_segmentation_cb_t segmentation_cb;
 	void				*data;
 	int				flags;
 	int				reconnect_timeout;
 	struct osmo_sock_init2_multiaddr_pars ma_pars;
 };
 
-void osmo_stream_cli_close(struct osmo_stream_cli *cli);
-
-/*! \brief Re-connect an Osmocom Stream Client
- *  If re-connection is enabled for this client
- *  (which is the case unless negative timeout was explicitly set via osmo_stream_cli_set_reconnect_timeout() call),
- *  we close any existing connection (if any) and schedule a re-connect timer */
-void osmo_stream_cli_reconnect(struct osmo_stream_cli *cli)
-{
-	osmo_stream_cli_close(cli);
-
-	if (cli->reconnect_timeout < 0) {
-		LOGSCLI(cli, LOGL_INFO, "not reconnecting, disabled\n");
-		return;
-	}
-
-	cli->state = STREAM_CLI_STATE_WAIT_RECONNECT;
-	LOGSCLI(cli, LOGL_INFO, "retrying reconnect in %d seconds...\n",
-		cli->reconnect_timeout);
-	osmo_timer_schedule(&cli->timer, cli->reconnect_timeout, 0);
-}
-
-/*! \brief Check if Osmocom Stream Client is in connected state
- *  \param[in] cli Osmocom Stream Client
- *  \return true if connected, false otherwise
+/*! \addtogroup stream_cli
+ *  @{
  */
-bool osmo_stream_cli_is_connected(struct osmo_stream_cli *cli)
-{
-	return cli->state == STREAM_CLI_STATE_CONNECTED;
-}
 
-static void osmo_stream_cli_close_iofd(struct osmo_stream_cli *cli)
+static void stream_cli_close_iofd(struct osmo_stream_cli *cli)
 {
 	if (!cli->iofd)
 		return;
 
-	osmo_iofd_close(cli->iofd);
+	osmo_iofd_free(cli->iofd);
+	cli->iofd = NULL;
 }
 
-static void osmo_stream_cli_close_ofd(struct osmo_stream_cli *cli)
+static void stream_cli_close_ofd(struct osmo_stream_cli *cli)
 {
 	if (cli->ofd.fd == -1)
 		return;
@@ -173,7 +138,7 @@ static void osmo_stream_cli_close_ofd(struct osmo_stream_cli *cli)
 	cli->ofd.fd = -1;
 }
 
-/*! \brief Close an Osmocom Stream Client
+/*! Close an Osmocom Stream Client.
  *  \param[in] cli Osmocom Stream Client to be closed
  *  We unregister the socket fd from the osmocom select() loop
  *  abstraction and close the socket */
@@ -192,10 +157,10 @@ void osmo_stream_cli_close(struct osmo_stream_cli *cli)
 
 	switch (cli->mode) {
 	case OSMO_STREAM_MODE_OSMO_FD:
-		osmo_stream_cli_close_ofd(cli);
+		stream_cli_close_ofd(cli);
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
-		osmo_stream_cli_close_iofd(cli);
+		stream_cli_close_iofd(cli);
 		break;
 	default:
 		OSMO_ASSERT(false);
@@ -203,14 +168,56 @@ void osmo_stream_cli_close(struct osmo_stream_cli *cli)
 
 	cli->state = STREAM_CLI_STATE_CLOSED;
 
-	if (old_state == STREAM_CLI_STATE_CONNECTED) {
+	/* If conn was established, notify the disconnection to the user:
+	 * Also, if reconnect is disabled by user, notify the user that connect() failed: */
+	if (old_state == STREAM_CLI_STATE_CONNECTED ||
+	    (old_state == STREAM_CLI_STATE_CONNECTING && cli->reconnect_timeout < 0)) {
 		LOGSCLI(cli, LOGL_DEBUG, "connection closed\n");
 		if (cli->disconnect_cb)
 			cli->disconnect_cb(cli);
 	}
 }
 
-/*! \brief Get file descriptor of the stream client socket
+/*! Re-connect an Osmocom Stream Client.
+ *  If re-connection is enabled for this client
+ *  (which is the case unless negative timeout was explicitly set via osmo_stream_cli_set_reconnect_timeout() call),
+ *  we close any existing connection (if any) and schedule a re-connect timer */
+void osmo_stream_cli_reconnect(struct osmo_stream_cli *cli)
+{
+	osmo_stream_cli_close(cli);
+
+	if (cli->reconnect_timeout < 0) {
+		LOGSCLI(cli, LOGL_INFO, "not reconnecting, disabled\n");
+		return;
+	}
+
+	cli->state = STREAM_CLI_STATE_WAIT_RECONNECT;
+	LOGSCLI(cli, LOGL_INFO, "retrying reconnect in %d seconds...\n",
+		cli->reconnect_timeout);
+	osmo_timer_schedule(&cli->timer, cli->reconnect_timeout, 0);
+}
+
+/*! Check if Osmocom Stream Client is in connected state.
+ *  \param[in] cli Osmocom Stream Client
+ *  \return true if connected, false otherwise
+ */
+bool osmo_stream_cli_is_connected(struct osmo_stream_cli *cli)
+{
+	return cli->state == STREAM_CLI_STATE_CONNECTED;
+}
+
+/*! Check if Osmocom Stream Client is opened (has an FD available) according to
+ *  its current state.
+ *  \param[in] cli Osmocom Stream Client
+ *  \return true if fd is available (osmo_stream_cli_get_fd()), false otherwise
+ */
+static bool stream_cli_is_opened(const struct osmo_stream_cli *cli)
+{
+	return cli->state == STREAM_CLI_STATE_CONNECTING ||
+	       cli->state == STREAM_CLI_STATE_CONNECTED;
+}
+
+/*! Retrieve file descriptor of the stream client socket.
  *  \param[in] cli Stream Client of which we want to obtain the file descriptor
  *  \returns File descriptor or negative in case of error */
 int
@@ -228,7 +235,19 @@ osmo_stream_cli_get_fd(const struct osmo_stream_cli *cli)
 	return -EINVAL;
 }
 
-static void osmo_stream_cli_read(struct osmo_stream_cli *cli)
+/*! Retrieve osmo_io descriptor of the stream client socket.
+ *  This function must not be called on a stream client in legacy osmo_fd mode!
+ *  The iofd is only valid once/after osmo_stream_cli_open() has successfully returned.
+ *  \param[in] cli Stream Client of which we want to obtain the file descriptor
+ *  \returns osmo_io_fd of stream client, or NULL if stream not yet opened. */
+struct osmo_io_fd *
+osmo_stream_cli_get_iofd(const struct osmo_stream_cli *cli)
+{
+	OSMO_ASSERT(cli->mode == OSMO_STREAM_MODE_OSMO_IO);
+	return cli->iofd;
+}
+
+static void stream_cli_read(struct osmo_stream_cli *cli)
 {
 	LOGSCLI(cli, LOGL_DEBUG, "message received\n");
 
@@ -236,7 +255,7 @@ static void osmo_stream_cli_read(struct osmo_stream_cli *cli)
 		cli->read_cb(cli);
 }
 
-static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
+static int stream_cli_write(struct osmo_stream_cli *cli)
 {
 #ifdef HAVE_LIBSCTP
 	struct sctp_sndrcvinfo sinfo;
@@ -303,6 +322,7 @@ static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
 		}
 		msgb_free(msg);
 		osmo_stream_cli_reconnect(cli);
+		return 0;
 	}
 
 	msgb_free(msg);
@@ -336,11 +356,13 @@ static void stream_cli_handle_connecting(struct osmo_stream_cli *cli, int res)
 	OSMO_ASSERT(fd >= 0);
 
 	if (ret < 0) {
+		LOGSCLI(cli, LOGL_ERROR, "connect failed (%d)\n", res);
 		osmo_stream_cli_reconnect(cli);
 		return;
 	}
 	ret = getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len);
 	if (ret >= 0 && error > 0) {
+		LOGSCLI(cli, LOGL_ERROR, "connect so_error (%d)\n", error);
 		osmo_stream_cli_reconnect(cli);
 		return;
 	}
@@ -385,11 +407,11 @@ static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
 	case STREAM_CLI_STATE_CONNECTED:
 		if (what & OSMO_FD_READ) {
 			LOGSCLI(cli, LOGL_DEBUG, "connected read\n");
-			osmo_stream_cli_read(cli);
+			stream_cli_read(cli);
 		}
 		if (what & OSMO_FD_WRITE) {
 			LOGSCLI(cli, LOGL_DEBUG, "connected write\n");
-			osmo_stream_cli_write(cli);
+			stream_cli_write(cli);
 		}
 		break;
 	default:
@@ -402,7 +424,7 @@ static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
 
 static void cli_timer_cb(void *data);
 
-/*! \brief Create an Osmocom stream client
+/*! Create an Osmocom stream client.
  *  \param[in] ctx talloc context from which to allocate memory
  *  This function allocates a new \ref osmo_stream_cli and initializes
  *  it with default values (5s reconnect timer, TCP protocol)
@@ -438,13 +460,29 @@ static void stream_cli_iofd_read_cb(struct osmo_io_fd *iofd, int res, struct msg
 
 	switch (cli->state) {
 	case STREAM_CLI_STATE_CONNECTING:
+		msgb_free(msg);
 		stream_cli_handle_connecting(cli, res);
 		break;
 	case STREAM_CLI_STATE_CONNECTED:
-		if (res == 0)
+		switch (res) {
+		case -EPIPE:
+		case -ECONNRESET:
+			LOGSCLI(cli, LOGL_ERROR, "lost connection with srv (%d)\n", res);
 			osmo_stream_cli_reconnect(cli);
-		else if (cli->iofd_read_cb)
-			cli->iofd_read_cb(cli, msg);
+			break;
+		case 0:
+			LOGSCLI(cli, LOGL_NOTICE, "connection closed with srv\n");
+			osmo_stream_cli_reconnect(cli);
+			break;
+		default:
+			LOGSCLI(cli, LOGL_DEBUG, "received %d bytes from srv\n", res);
+			break;
+		}
+		/* Notify user of new data or error: */
+		if (cli->iofd_read_cb)
+			cli->iofd_read_cb(cli, res, msg);
+		else
+			msgb_free(msg);
 		break;
 	default:
 		osmo_panic("%s() called with unexpected state %d\n", __func__, cli->state);
@@ -470,14 +508,60 @@ static void stream_cli_iofd_write_cb(struct osmo_io_fd *iofd, int res, struct ms
 	}
 }
 
-static struct osmo_io_ops osmo_stream_cli_ioops = {
+static const struct osmo_io_ops osmo_stream_cli_ioops = {
 	.read_cb = stream_cli_iofd_read_cb,
 	.write_cb = stream_cli_iofd_write_cb,
 
 	.segmentation_cb = NULL,
 };
 
-/*! \brief Set a name on the cli object (used during logging)
+#ifdef HAVE_LIBSCTP
+static void stream_cli_iofd_recvmsg_cb(struct osmo_io_fd *iofd, int res, struct msgb *msg, const struct msghdr *msgh)
+{
+	struct osmo_stream_cli *cli  = osmo_iofd_get_data(iofd);
+
+	res = stream_iofd_sctp_recvmsg_trailer(iofd, msg, res, msgh);
+
+	switch (cli->state) {
+	case STREAM_CLI_STATE_CONNECTING:
+		msgb_free(msg);
+		stream_cli_handle_connecting(cli, res);
+		break;
+	case STREAM_CLI_STATE_CONNECTED:
+		switch (res) {
+		case -EPIPE:
+		case -ECONNRESET:
+			LOGSCLI(cli, LOGL_ERROR, "lost connection with srv (%d)\n", res);
+			osmo_stream_cli_reconnect(cli);
+			break;
+		case 0:
+			LOGSCLI(cli, LOGL_NOTICE, "connection closed with srv\n");
+			osmo_stream_cli_reconnect(cli);
+			break;
+		default:
+			break;
+		}
+		/* Notify user of new data or error: */
+		if (cli->iofd_read_cb)
+			cli->iofd_read_cb(cli, res, msg);
+		else
+			msgb_free(msg);
+		break;
+	default:
+		osmo_panic("%s() called with unexpected state %d\n", __func__, cli->state);
+	}
+}
+
+static const struct osmo_io_ops osmo_stream_cli_ioops_sctp = {
+	.recvmsg_cb = stream_cli_iofd_recvmsg_cb,
+	.sendmsg_cb = stream_cli_iofd_write_cb,
+
+	.segmentation_cb = NULL,
+};
+#endif
+
+
+/*! Set a name on the cli object (used during logging).
  *  \param[in] cli stream_cli whose name is to be set
  *  \param[in] name the name to be set on cli
  */
@@ -488,7 +572,17 @@ void osmo_stream_cli_set_name(struct osmo_stream_cli *cli, const char *name)
 		osmo_iofd_set_name(cli->iofd, name);
 }
 
-/*! \brief Set the remote address to which we connect
+/*! Retrieve name previously set on the cli object (see osmo_stream_cli_set_name()).
+ *  \param[in] cli stream_cli whose name is to be retrieved
+ *  \returns The name to be set on cli; NULL if never set
+ */
+const char *osmo_stream_cli_get_name(const struct osmo_stream_cli *cli)
+{
+	return cli->name;
+}
+
+/*! Set the remote address to which we connect.
+ *  Any changes to this setting will only become active upon next (re)connect.
  *  \param[in] cli Stream Client to modify
  *  \param[in] addr Remote IP address
  */
@@ -498,8 +592,9 @@ osmo_stream_cli_set_addr(struct osmo_stream_cli *cli, const char *addr)
 	osmo_stream_cli_set_addrs(cli, &addr, 1);
 }
 
-/*! \brief Set the remote address set to which we connect.
+/*! Set the remote address set to which we connect.
  *  Useful for protocols allowing connecting to more than one address (such as SCTP)
+ *  Any changes to this setting will only become active upon next (re)connect.
  *  \param[in] cli Stream Client to modify
  *  \param[in] addr Remote IP address set
  *  \return negative on error, 0 on success
@@ -523,7 +618,8 @@ int osmo_stream_cli_set_addrs(struct osmo_stream_cli *cli, const char **addr, si
 	return 0;
 }
 
-/*! \brief Set the remote port number to which we connect
+/*! Set the remote port number to which we connect.
+ *  Any changes to this setting will only become active upon next (re)connect.
  *  \param[in] cli Stream Client to modify
  *  \param[in] port Remote port number
  */
@@ -534,7 +630,8 @@ osmo_stream_cli_set_port(struct osmo_stream_cli *cli, uint16_t port)
 	cli->flags |= OSMO_STREAM_CLI_F_RECONF;
 }
 
-/*! \brief Set the local port number for the socket (to be bound to)
+/*! Set the local port number for the socket (to be bound to).
+ *  Any changes to this setting will only become active upon next (re)connect.
  *  \param[in] cli Stream Client to modify
  *  \param[in] port Local port number
  */
@@ -545,7 +642,8 @@ osmo_stream_cli_set_local_port(struct osmo_stream_cli *cli, uint16_t port)
 	cli->flags |= OSMO_STREAM_CLI_F_RECONF;
 }
 
-/*! \brief Set the local address for the socket (to be bound to)
+/*! Set the local address for the socket (to be bound to).
+ *  Any changes to this setting will only become active upon next (re)connect.
  *  \param[in] cli Stream Client to modify
  *  \param[in] port Local host name
  */
@@ -555,8 +653,9 @@ osmo_stream_cli_set_local_addr(struct osmo_stream_cli *cli, const char *addr)
 	osmo_stream_cli_set_local_addrs(cli, &addr, 1);
 }
 
-/*! \brief Set the local address set to which we connect.
+/*! Set the local address set to which we bind.
  *  Useful for protocols allowing bind to more than one address (such as SCTP)
+ *  Any changes to this setting will only become active upon next (re)connect.
  *  \param[in] cli Stream Client to modify
  *  \param[in] addr Local IP address set
  *  \return negative on error, 0 on success
@@ -580,7 +679,8 @@ int osmo_stream_cli_set_local_addrs(struct osmo_stream_cli *cli, const char **ad
 	return 0;
 }
 
-/*! \brief Set the protocol for the stream client socket
+/*! Set the protocol for the stream client socket.
+ *  Any changes to this setting will only become active upon next (re)connect.
  *  \param[in] cli Stream Client to modify
  *  \param[in] proto Protocol (like IPPROTO_TCP (default), IPPROTO_SCTP, ...)
  */
@@ -592,32 +692,34 @@ osmo_stream_cli_set_proto(struct osmo_stream_cli *cli, uint16_t proto)
 }
 
 /* Configure client side segmentation for the iofd */
-static void configure_cli_segmentation_cb(struct osmo_io_fd *iofd,
-					       int (*segmentation_cb)(struct msgb *msg))
+static void configure_cli_segmentation_cb(struct osmo_stream_cli *cli,
+					  osmo_stream_cli_segmentation_cb_t segmentation_cb)
 {
 	/* Copy default settings */
-	struct osmo_io_ops client_ops = osmo_stream_cli_ioops;
+	struct osmo_io_ops client_ops;
+	osmo_iofd_get_ioops(cli->iofd, &client_ops);
 	/* Set segmentation cb for this client */
 	client_ops.segmentation_cb = segmentation_cb;
-	osmo_iofd_set_ioops(iofd, &client_ops);
+	osmo_iofd_set_ioops(cli->iofd, &client_ops);
 }
 
-/*! \brief Set the segmentation callback for the client
+/*! Set the segmentation callback for the client.
  *  \param[in,out] cli Stream Client to modify
  *  \param[in] segmentation_cb Target segmentation callback
  */
 void osmo_stream_cli_set_segmentation_cb(struct osmo_stream_cli *cli,
-					 int (*segmentation_cb)(struct msgb *msg))
+					 osmo_stream_cli_segmentation_cb_t segmentation_cb)
 {
 	cli->segmentation_cb = segmentation_cb;
 	if (cli->iofd) /* Otherwise, this will be done in osmo_stream_cli_open() */
-		configure_cli_segmentation_cb(cli->iofd, segmentation_cb);
+		configure_cli_segmentation_cb(cli, segmentation_cb);
 }
 
-/*! \brief Set the socket type for the stream server link
+/*! Set the socket type for the stream server link.
+ *  Any changes to this setting will only become active upon next (re)connect.
  *  \param[in] cli Stream Client to modify
  *  \param[in] type Socket Type (like SOCK_STREAM (default), SOCK_SEQPACKET, ...)
- *  \returns zero on success, negative on error.
+ *  \returns zero on success, negative -errno on error.
  */
 int osmo_stream_cli_set_type(struct osmo_stream_cli *cli, int type)
 {
@@ -633,10 +735,11 @@ int osmo_stream_cli_set_type(struct osmo_stream_cli *cli, int type)
 	return 0;
 }
 
-/*! \brief Set the socket type for the stream server link
+/*! Set the socket type for the stream server link.
+ *  Any changes to this setting will only become active upon next (re)connect.
  *  \param[in] cli Stream Client to modify
  *  \param[in] type Socket Domain (like AF_UNSPEC (default for IP), AF_UNIX, AF_INET, ...)
- *  \returns zero on success, negative on error.
+ *  \returns zero on success, negative -errno on error.
  */
 int osmo_stream_cli_set_domain(struct osmo_stream_cli *cli, int domain)
 {
@@ -654,7 +757,7 @@ int osmo_stream_cli_set_domain(struct osmo_stream_cli *cli, int domain)
 	return 0;
 }
 
-/*! \brief Set the reconnect time of the stream client socket
+/*! Set the reconnect time of the stream client socket.
  *  \param[in] cli Stream Client to modify
  *  \param[in] timeout Re-connect timeout in seconds or negative value to disable auto-reconnection */
 void
@@ -663,7 +766,7 @@ osmo_stream_cli_set_reconnect_timeout(struct osmo_stream_cli *cli, int timeout)
 	cli->reconnect_timeout = timeout;
 }
 
-/*! \brief Set application private data of the stream client socket
+/*! Set application private data of the stream client socket.
  *  \param[in] cli Stream Client to modify
  *  \param[in] data User-specific data (available in call-back functions) */
 void
@@ -672,7 +775,7 @@ osmo_stream_cli_set_data(struct osmo_stream_cli *cli, void *data)
 	cli->data = data;
 }
 
-/*! \brief Get application private data of the stream client socket
+/*! Retrieve application private data of the stream client socket.
  *  \param[in] cli Stream Client to modify
  *  \returns Application private data, as set by \ref osmo_stream_cli_set_data() */
 void *osmo_stream_cli_get_data(struct osmo_stream_cli *cli)
@@ -680,7 +783,9 @@ void *osmo_stream_cli_get_data(struct osmo_stream_cli *cli)
 	return cli->data;
 }
 
-/*! \brief Get the stream client socket description.
+/*! Retrieve the stream client socket description.
+ *  Calling this function will build a string that describes the socket in terms of its local/remote
+ *  address/port.  The returned name is stored in a static buffer; it is hence not re-entrant or thread-safe.
  *  \param[in] cli Stream Client to examine
  *  \returns Socket description or NULL in case of error */
 char *osmo_stream_cli_get_sockname(const struct osmo_stream_cli *cli)
@@ -693,7 +798,8 @@ char *osmo_stream_cli_get_sockname(const struct osmo_stream_cli *cli)
 	return buf;
 }
 
-/*! \brief Get Osmocom File Descriptor of the stream client socket
+/*! Retrieve Osmocom File Descriptor of the stream client socket.
+ *  This function only works in case you operate osmo_stream_cli in osmo_fd mode!
  *  \param[in] cli Stream Client to modify
  *  \returns Pointer to \ref osmo_fd */
 struct osmo_fd *
@@ -703,52 +809,54 @@ osmo_stream_cli_get_ofd(struct osmo_stream_cli *cli)
 	return &cli->ofd;
 }
 
-/*! \brief Set the call-back function called on connect of the stream client socket
+/*! Set the call-back function called on connect of the stream client socket.
+ *  The call-back function registered via this function will be called upon completion of the non-blocking
+ *  outbound connect operation.
  *  \param[in] cli Stream Client to modify
  *  \param[in] connect_cb Call-back function to be called upon connect */
 void
 osmo_stream_cli_set_connect_cb(struct osmo_stream_cli *cli,
-	int (*connect_cb)(struct osmo_stream_cli *cli))
+			       osmo_stream_cli_connect_cb_t connect_cb)
 {
 	cli->connect_cb = connect_cb;
 }
 
-/*! \brief Set the call-back function called on disconnect of the stream client socket
+/*! Set the call-back function called on disconnect of the stream client socket.
  *  \param[in] cli Stream Client to modify
  *  \param[in] disconnect_cb Call-back function to be called upon disconnect */
 void osmo_stream_cli_set_disconnect_cb(struct osmo_stream_cli *cli,
-				       int (*disconnect_cb)(struct osmo_stream_cli *cli))
+				       osmo_stream_cli_disconnect_cb_t disconnect_cb)
 {
 	cli->disconnect_cb = disconnect_cb;
 }
 
-/*! \brief Set the call-back function called to read from the stream client socket
- *  This function will configure osmo_stream_cli to use osmo_ofd internally.
+/*! Set the call-back function called to read from the stream client socket.
+ *  This function will implicitly configure osmo_stream_cli to use legacy osmo_ofd mode.
  *  \param[in] cli Stream Client to modify
  *  \param[in] read_cb Call-back function to be called when we want to read */
 void
 osmo_stream_cli_set_read_cb(struct osmo_stream_cli *cli,
-			    int (*read_cb)(struct osmo_stream_cli *cli))
+			    osmo_stream_cli_read_cb_t read_cb)
 {
 	OSMO_ASSERT(cli->mode != OSMO_STREAM_MODE_OSMO_IO);
 	cli->mode = OSMO_STREAM_MODE_OSMO_FD;
 	cli->read_cb = read_cb;
 }
 
-/*! \brief Set the call-back function called to read from the stream client socket
- *  This function will configure osmo_stream_cli to use osmo_iofd internally.
+/*! Set the call-back function called to read from the stream client socket.
+ *  This function will implicitly configure osmo_stream_cli to use osmo_iofd mode.
  *  \param[in] cli Stream Client to modify
  *  \param[in] read_cb Call-back function to be called when data was read from the socket */
 void
 osmo_stream_cli_set_read_cb2(struct osmo_stream_cli *cli,
-			    int (*read_cb)(struct osmo_stream_cli *cli, struct msgb *msg))
+			     osmo_stream_cli_read_cb2_t read_cb)
 {
 	OSMO_ASSERT(cli->mode != OSMO_STREAM_MODE_OSMO_FD);
 	cli->mode = OSMO_STREAM_MODE_OSMO_IO;
 	cli->iofd_read_cb = read_cb;
 }
 
-/*! \brief Destroy a Osmocom stream client (includes close)
+/*! Destroy a Osmocom stream client (includes close).
  *  \param[in] cli Stream Client to destroy */
 void osmo_stream_cli_destroy(struct osmo_stream_cli *cli)
 {
@@ -758,7 +866,7 @@ void osmo_stream_cli_destroy(struct osmo_stream_cli *cli)
 	talloc_free(cli);
 }
 
-/*! \brief DEPRECATED: use osmo_stream_cli_set_reconnect_timeout() or osmo_stream_cli_reconnect() instead!
+/*! DEPRECATED: use osmo_stream_cli_set_reconnect_timeout() or osmo_stream_cli_reconnect() instead!
  * Open connection of an Osmocom stream client
  *  \param[in] cli Stream Client to connect
  *  \param[in] reconect 1 if we should not automatically reconnect
@@ -780,7 +888,8 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 		ret = osmo_sock_init2_multiaddr2(AF_UNSPEC, SOCK_STREAM, cli->proto,
 						(const char **)cli->local_addr, cli->local_addrcnt, cli->local_port,
 						(const char **)cli->addr, cli->addrcnt, cli->port,
-						OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK,
+						OSMO_SOCK_F_CONNECT | OSMO_SOCK_F_BIND | OSMO_SOCK_F_NONBLOCK |
+						OSMO_SOCK_F_DSCP(cli->ip_dscp) | OSMO_SOCK_F_PRIO(cli->sk_prio),
 						&cli->ma_pars);
 		break;
 #endif
@@ -788,10 +897,12 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 		ret = osmo_sock_init2(AF_UNSPEC, SOCK_STREAM, cli->proto,
 				      cli->local_addr[0], cli->local_port,
 				      cli->addr[0], cli->port,
-				      OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+				      OSMO_SOCK_F_CONNECT | OSMO_SOCK_F_BIND | OSMO_SOCK_F_NONBLOCK |
+				      OSMO_SOCK_F_DSCP(cli->ip_dscp) | OSMO_SOCK_F_PRIO(cli->sk_prio));
 	}
 
 	if (ret < 0) {
+		LOGSCLI(cli, LOGL_ERROR, "connect: socket creation error (%d)\n", ret);
 		if (reconnect)
 			osmo_stream_cli_reconnect(cli);
 		return ret;
@@ -816,24 +927,100 @@ error_close_socket:
 	return -EIO;
 }
 
-/*! \brief Set the NODELAY socket option to avoid Nagle-like behavior
+/*! Set the NODELAY socket option to avoid Nagle-like behavior.
  *  Setting this to nodelay=true will automatically set the NODELAY
  *  socket option on any socket established via \ref osmo_stream_cli_open
- *  or any re-connect.  You have to set this _before_ opening the
+ *  or any re-connect. This can be set either before or after opening the
  *  socket.
  *  \param[in] cli Stream client whose sockets are to be configured
  *  \param[in] nodelay whether to set (true) NODELAY before connect()
  */
 void osmo_stream_cli_set_nodelay(struct osmo_stream_cli *cli, bool nodelay)
 {
+	int fd;
 	if (nodelay)
 		cli->flags |= OSMO_STREAM_CLI_F_NODELAY;
 	else
 		cli->flags &= ~OSMO_STREAM_CLI_F_NODELAY;
+
+	if (!stream_cli_is_opened(cli))
+		return; /* Config will be applied upon open() time */
+
+	if ((fd = osmo_stream_cli_get_fd(cli)) < 0) {
+		LOGSCLI(cli, LOGL_ERROR, "set_nodelay(%u): failed obtaining socket\n", nodelay);
+		return;
+	}
+	if (stream_setsockopt_nodelay(fd, cli->proto, nodelay ? 1 : 0) < 0)
+		LOGSCLI(cli, LOGL_ERROR, "set_nodelay(%u): failed setsockopt err=%d\n",
+			nodelay, errno);
 }
 
-/*! \brief Open connection of an Osmocom stream client
- *  By default the client will automatically reconnect after default timeout.
+/*! Set the priority value of the stream socket.
+ *  Setting this  will automatically set the socket priority
+ *  option on any socket established via \ref osmo_stream_cli_open
+ *  or any re-connect. This can be set either before or after opening the
+ *  socket.
+ *  \param[in] cli Stream client whose sockets are to be configured
+ *  \param[in] sk_prio priority value. Values outside 0..6 require CAP_NET_ADMIN.
+ *  \return negative on error, 0 on success
+ */
+int osmo_stream_cli_set_priority(struct osmo_stream_cli *cli, int sk_prio)
+{
+	int rc;
+	int fd;
+
+	if (cli->sk_prio == sk_prio)
+		return 0; /* No change needed */
+
+	cli->sk_prio = sk_prio;
+
+	if (!stream_cli_is_opened(cli))
+		return 0; /* Config will be applied upon open() time */
+
+	if ((fd = osmo_stream_cli_get_fd(cli)) < 0) { /* Shouldn't happen... */
+		LOGSCLI(cli, LOGL_ERROR, "set_priority(%d): failed obtaining socket\n", cli->sk_prio);
+		return -EBADFD;
+	}
+	if ((rc = osmo_sock_set_priority(fd, cli->sk_prio)) < 0)
+		LOGSCLI(cli, LOGL_ERROR, "set_priority(%d): failed setsockopt err=%d\n",
+			cli->sk_prio, errno);
+	return rc;
+}
+
+/*! Set the DSCP (differentiated services code point) of the stream socket.
+ *  Setting this  will automatically set the IP DSCP option on any socket established
+ *  via \ref osmo_stream_cli_open or any re-connect. This can be set either before or
+ *  after opening the socket.
+ *  \param[in] cli Stream client whose sockets are to be configured
+ *  \param[in] ip_dscp DSCP value. Value range 0..63.
+ *  \return negative on error, 0 on success
+ */
+int osmo_stream_cli_set_ip_dscp(struct osmo_stream_cli *cli, uint8_t ip_dscp)
+{
+	int rc;
+	int fd;
+
+	if (cli->ip_dscp == ip_dscp)
+		return 0; /* No change needed */
+
+	cli->ip_dscp = ip_dscp;
+
+	if (!stream_cli_is_opened(cli))
+		return 0; /* Config will be applied upon open() time */
+
+	if ((fd = osmo_stream_cli_get_fd(cli)) < 0) { /* Shouldn't happen... */
+		LOGSCLI(cli, LOGL_ERROR, "set_ip_dscp(%u): failed obtaining socket\n", cli->ip_dscp);
+		return -EBADFD;
+	}
+	if ((rc = osmo_sock_set_dscp(fd, cli->ip_dscp)) < 0)
+		LOGSCLI(cli, LOGL_ERROR, "set_ip_dscp(%u): failed setsockopt err=%d\n",
+			cli->ip_dscp, errno);
+	return rc;
+}
+
+/*! Open connection of an Osmocom stream client.
+ *  This will initiate an non-blocking outbound connect to the configured destination (server) address.
+ *  By default the client will automatically attempt to reconnect after default timeout.
  *  To disable this, use osmo_stream_cli_set_reconnect_timeout() before calling this function.
  *  \param[in] cli Stream Client to connect
  *  \return negative on error, 0 on success */
@@ -860,7 +1047,8 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 #ifdef HAVE_LIBSCTP
 		case IPPROTO_SCTP:
 			local_addrcnt = cli->local_addrcnt;
-			flags = OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_NONBLOCK;
+			flags = OSMO_SOCK_F_CONNECT | OSMO_SOCK_F_NONBLOCK |
+				OSMO_SOCK_F_DSCP(cli->ip_dscp) | OSMO_SOCK_F_PRIO(cli->sk_prio);
 			if (cli->local_addrcnt > 0 || cli->local_port > 0) { /* explicit bind required? */
 				flags |= OSMO_SOCK_F_BIND;
 				/* If no local addr configured, use local_addr[0]=NULL by default when creating the socket. */
@@ -877,7 +1065,8 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 			ret = osmo_sock_init2(cli->sk_domain, cli->sk_type, cli->proto,
 					      cli->local_addr[0], cli->local_port,
 					      cli->addr[0], cli->port,
-					      OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+					      OSMO_SOCK_F_CONNECT | OSMO_SOCK_F_BIND | OSMO_SOCK_F_NONBLOCK |
+					      OSMO_SOCK_F_DSCP(cli->ip_dscp) | OSMO_SOCK_F_PRIO(cli->sk_prio));
 		}
 		break;
 	default:
@@ -885,6 +1074,7 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 	}
 
 	if (ret < 0) {
+		LOGSCLI(cli, LOGL_ERROR, "connect: socket creation error (%d)\n", ret);
 		osmo_stream_cli_reconnect(cli);
 		return ret;
 	}
@@ -904,15 +1094,30 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 			goto error_close_socket;
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
-		if (!cli->iofd)
-			cli->iofd = osmo_iofd_setup(cli, fd, cli->name, OSMO_IO_FD_MODE_READ_WRITE, &osmo_stream_cli_ioops, cli);
+		/* Be sure that previous osmo_io instance is freed before creating a new one. */
+		stream_cli_close_iofd(cli);
+#ifdef HAVE_LIBSCTP
+		if (cli->proto == IPPROTO_SCTP) {
+			cli->iofd = osmo_iofd_setup(cli, fd, cli->name, OSMO_IO_FD_MODE_RECVMSG_SENDMSG,
+						    &osmo_stream_cli_ioops_sctp, cli);
+			if (cli->iofd)
+				osmo_iofd_set_cmsg_size(cli->iofd, CMSG_SPACE(sizeof(struct sctp_sndrcvinfo)));
+		} else {
+#else
+		if (true) {
+#endif
+			cli->iofd = osmo_iofd_setup(cli, fd, cli->name, OSMO_IO_FD_MODE_READ_WRITE,
+						    &osmo_stream_cli_ioops, cli);
+		}
 		if (!cli->iofd)
 			goto error_close_socket;
-		configure_cli_segmentation_cb(cli->iofd, cli->segmentation_cb);
+
+		osmo_iofd_notify_connected(cli->iofd);
+
+		configure_cli_segmentation_cb(cli, cli->segmentation_cb);
 
 		if (osmo_iofd_register(cli->iofd, fd) < 0)
 			goto error_close_socket;
-		osmo_iofd_notify_connected(cli->iofd);
 		break;
 	default:
 		OSMO_ASSERT(false);
@@ -937,32 +1142,52 @@ static void cli_timer_cb(void *data)
 	osmo_stream_cli_open(cli);
 }
 
-/*! \brief Enqueue data to be sent via an Osmocom stream client
+/*! Enqueue data to be sent via an Osmocom stream client..
+ *  This is the function you use for writing/sending/transmitting data via the osmo_stream_cli.
  *  \param[in] cli Stream Client through which we want to send
  *  \param[in] msg Message buffer to enqueue in transmit queue */
 void osmo_stream_cli_send(struct osmo_stream_cli *cli, struct msgb *msg)
 {
+	int rc;
+
 	OSMO_ASSERT(cli);
 	OSMO_ASSERT(msg);
 
+	if (!osmo_stream_cli_is_connected(cli)) {
+		LOGSCLI(cli, LOGL_ERROR, "send: not connected, dropping data!\n");
+		msgb_free(msg);
+		return;
+	}
+
 	switch (cli->mode) {
 	case OSMO_STREAM_MODE_OSMO_FD:
 		msgb_enqueue(&cli->tx_queue, msg);
 		osmo_fd_write_enable(&cli->ofd);
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
-		osmo_iofd_write_msgb(cli->iofd, msg);
+		/* whenever osmo_stream_cli_is_connected() [see above check], we should have an iofd */
+		OSMO_ASSERT(cli->iofd);
+		if (cli->proto == IPPROTO_SCTP)
+			rc = stream_iofd_sctp_send_msgb(cli->iofd, msg, MSG_NOSIGNAL);
+		else
+			rc = osmo_iofd_write_msgb(cli->iofd, msg);
+		if (rc < 0)
+			msgb_free(msg);
 		break;
 	default:
 		OSMO_ASSERT(false);
 	}
 }
 
-/*! \brief Receive data via an Osmocom stream client
+/*! Receive data via an Osmocom stream client in osmo_fd mode.
  *  \param[in] cli Stream Client through which we want to send
  *  \param msg pre-allocate message buffer to which received data is appended
  *  \returns number of bytes read; <=0 in case of error
  *
+ *  Application programs using the legacy osmo_fd mode of osmo_stream_cli will use
+ *  this function to read/receive from a stream client socket after they have been notified that
+ *  it is readable (via select/poll).
+ *
  *  If conn is an SCTP connection, additional specific considerations shall be taken:
  *  - msg->cb is always filled with SCTP ppid, and SCTP stream values, see msgb_sctp_*() APIs.
  *  - If an SCTP notification was received when reading from the SCTP socket,
@@ -978,6 +1203,7 @@ int osmo_stream_cli_recv(struct osmo_stream_cli *cli, struct msgb *msg)
 	int ret;
 	OSMO_ASSERT(cli);
 	OSMO_ASSERT(msg);
+	OSMO_ASSERT(cli->mode == OSMO_STREAM_MODE_OSMO_FD);
 
 	switch (cli->sk_domain) {
 	case AF_UNIX:
@@ -1007,10 +1233,12 @@ int osmo_stream_cli_recv(struct osmo_stream_cli *cli, struct msgb *msg)
 	}
 
 	if (ret < 0) {
-		if (ret == -EAGAIN)
+		if (ret == -EAGAIN) /* Received MSG_NOTIFICATION from stream_sctp_recvmsg_wrapper() */
 			return ret;
 		if (errno == EPIPE || errno == ECONNRESET)
-			LOGSCLI(cli, LOGL_ERROR, "lost connection with srv\n");
+			LOGSCLI(cli, LOGL_ERROR, "lost connection with srv (%d)\n", errno);
+		else
+			LOGSCLI(cli, LOGL_ERROR, "recv failed (%d)\n", errno);
 		osmo_stream_cli_reconnect(cli);
 		return ret;
 	} else if (ret == 0) {
@@ -1023,6 +1251,8 @@ int osmo_stream_cli_recv(struct osmo_stream_cli *cli, struct msgb *msg)
 	return ret;
 }
 
+/*! Clear the transmit queue of the stream client.
+ *  Calling this function wil clear (delete) any pending, not-yet transmitted data from the transmit queue. */
 void osmo_stream_cli_clear_tx_queue(struct osmo_stream_cli *cli)
 {
 	switch (cli->mode) {
@@ -1041,6 +1271,12 @@ void osmo_stream_cli_clear_tx_queue(struct osmo_stream_cli *cli)
 	}
 }
 
+/*! Set given parameter of stream client to given value.
+ *  \param[in] cli stream client on which to set parameter.
+ *  \param[in] par identifier of the parameter to be set.
+ *  \param[in] val value of the parameter to be set.
+ *  \param[in] val_len length of the parameter value.
+ *  \returns 0 in success; negative -errno on error. */
 int osmo_stream_cli_set_param(struct osmo_stream_cli *cli, enum osmo_stream_cli_param par, void *val, size_t val_len)
 {
 	OSMO_ASSERT(cli);
